from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import User


class EmailBackend(BaseBackend):
    """
    Authentication backend to allow login by email.

    - authenticate() uses email and password credentials.
    """

    def get_user(self, user_id):
        try:
            user = User.objects.get(pk=user_id, is_active=True)
        except User.DoesNotExist:
            return None
        return user

    def authenticate(self, request, email=None, password=None, **kwargs):
        if email is None or password is None:
            return

        try:
            user = User.objects.get(email=email, is_active=True)
        except User.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a non-existing user. See ModelBackend.
            User().set_password(password)
        else:
            if user.check_password(password):
                return user



from django import forms
from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import TestCase
from unique_user_email.backend import EmailBackend
from unique_user_email.forms import AuthenticationForm
from django.db.models.constraints import UniqueConstraint


class UniqueEmailTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username="testuser",
            email="test@example.com",
            password="testpass",
        )

    def test_model_form_disallows_duplicate_emails(self):
        class UserForm(forms.ModelForm):
            class Meta:
                model = User
                fields = ["username", "email"]

        form = UserForm(
            data={
                "username": "testuser2",
                "email": self.user.email,
            }
        )
        self.assertIs(False, form.is_valid())

    def test_full_clean_disallows_duplicate_emails(self):
        user2 = User(
            username="testuser2",
            email=self.user.email,
            password="testpass2",
        )

        with self.assertRaisesMessage(
            ValidationError,
            "User with this Email address already exists.",
        ):
            user2.full_clean()

    def test_model_save_disallows_duplicate_emails(self):
        user2 = User(
            username="testuser2",
            email=self.user.email,
            password="testpass2",
        )
        with self.assertRaisesMessage(
            IntegrityError,
            "UNIQUE constraint failed: auth_user.email",
        ):
            user2.save()

    def test_user_constraints(self):
        self.assertIsInstance(User._meta.constraints[0], UniqueConstraint)
        self.assertEqual("unique_user_email", User._meta.constraints[0].name)
        self.assertEqual(
            "unique_user_email",
            User._meta.original_attrs.get("constraints")[0].name
        )

class EmailBackendTests(TestCase):
    def test_none_for_username_logins(self):
        """
        Authenticate with username and password returns None.
        """
        request = None
        backend = EmailBackend()
        self.assertIsNone(
            backend.authenticate(request, username="testUser", password="testPassword")
        )

    def test_user_can_login_with_email(self):
        """
        Authenticate with email and password returns the user.
        """
        user = User.objects.create_user(
            username="testUser", email="a@b.com", password="testPassword"
        )
        request = None
        backend = EmailBackend()
        authed_user = backend.authenticate(
            request, email="a@b.com", password="testPassword"
        )
        self.assertEqual(user, authed_user)

    def test_email_backend_is_installed(self):
        """
        EmailBackend is listed in settings.AUTHENTICATION_BACKENDS.
        """
        self.assertIn(
            "unique_user_email.backend.EmailBackend",
            settings.AUTHENTICATION_BACKENDS,
        )


class AuthenticationFormTests(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(
            username="testUser",
            email="a@b.com",
            password="testPassword",
            is_staff=True,
        )

    def test_get_credentials_with_email(self):
        """
        get_credentials() returns email kwarg.
        """
        form = AuthenticationForm(data={"login": "a@b.com", "password": "testPassword"})
        self.assertIs(True, form.is_valid(), form.errors)
        self.assertEqual(
            {"email": "a@b.com", "password": "testPassword"}, form.get_credentials()
        )

    def test_login_view(self):
        tests = ["testUser", "a@b.com"]
        for test in tests:
            with self.subTest(login=test):
                data = {"login": test, "password": "testPassword"}
                url = "/accounts/login/"
                response = self.client.post(url, data=data)
                self.assertRedirects(
                    response,
                    settings.LOGIN_REDIRECT_URL,
                )



from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth.forms import UsernameField
from django.core import exceptions, validators


class AuthenticationForm(forms.Form):
    """
    Form allowing login by username or email.

    - Takes login and password.
    - Maps to credentials for authentication, using email if login is an email
      address.
    - Exposes the user for use in the view.
    """

    login = UsernameField(max_length=150, label="Username or Email")
    password = forms.CharField(
        label="Password",
        widget=forms.PasswordInput(attrs={"autocomplete": "current-password"}),
    )

    def __init__(self, request=None, *args, **kwargs):
        self.request = request  # ???: Do we need this?
        self.user_cache = None
        super().__init__(*args, **kwargs)

    def get_credentials(self):
        """
        Maps the login to credentials for authentication.

        If the login is an email address, use the email kwarg, otherwise use
        the username kwarg.
        """
        credentials = {}
        login = self.cleaned_data.get("login")
        if self.is_email(login):
            credentials["email"] = login
        else:
            credentials["username"] = login
        credentials["password"] = self.cleaned_data.get("password")
        return credentials

    def is_email(self, login):
        try:
            validators.validate_email(login)
            is_email = True
        except exceptions.ValidationError:
            is_email = False
        return is_email

    def clean(self):
        super().clean()
        if self._errors:
            print(self._errors)
            return

        credentials = self.get_credentials()
        self.user_cache = authenticate(self.request, **credentials)
        if self.user_cache is None:
            raise forms.ValidationError(
                "Please enter a correct username/email and password. Note that both "
                "fields are case-sensitive."
            )
        return self.cleaned_data

    def get_user(self):
        return self.user_cache
